#ifdef HAVE_CONFIG_H
#include <linux/config.h>
#endif

#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/usb.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/kdev_t.h>
#include <linux/types.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <linux/rwsem.h>
#include <linux/uaccess.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/completion.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#include <linux/workqueue.h>
#include <linux/semaphore.h>

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a<<16)|(b<<8)|c)
#endif

#ifdef CONFIG_COMPAT
#include <linux/compat.h>
#endif

#include "wrg315ioctl.h"

#define PCI_BUFFER_SIZE     8192

#define MAX_WRG3_DEVICES    16
#define WiNRADiO_MAJOR      82
#define WRG3_MINOR_BASE     (32+32)
#define WRG3_MINOR_END      (WRG3_MINOR_BASE+MAX_WRG3_DEVICES-1)

#define WRG3_NAME "wrg315"

#define WRG3_PCI 0
#define WRG3_USB 1

#define WRG3_DISPLAY_NAME "WRG315"

#define MAX_PIC_RETRY 128

#define USB_TIMEOUT (2000)

#define MAX_ENDPOINTS 16

#define TO_JIFFIES(x) ((x*HZ)/1000)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
#define USE_MUTEX
#else
#undef USE_MUTEX
#endif


#ifdef USE_MUTEX

#include <linux/mutex.h>

#define WRG3_MUTEX struct mutex
#define WRG3_INIT_MUTEX(object)         mutex_init(&object)
#define WRG3_DESTROY_MUTEX(object)      mutex_destroy(&object)
#define WRG3_LOCK(object)               mutex_lock(&object)
#define WRG3_LOCK_INTERRUPTIBLE(object) mutex_lock_interruptible(&object)
#define WRG3_UNLOCK(object)             mutex_unlock(&object)

#else

#define WRG3_MUTEX struct semaphore
#define WRG3_INIT_MUTEX(object)         init_MUTEX(&object)
#define WRG3_DESTROY_MUTEX(object)
#define WRG3_LOCK(object)               down(&object)
#define WRG3_LOCK_INTERRUPTIBLE(object) down_interruptible(&object)
#define WRG3_UNLOCK(object)             up(&object)

#endif


typedef struct _ASYNC_BULK_URB
{
    unsigned int            SeqNum;
    struct urb             *Urb;
    struct __WRG3          *Dev;
    struct list_head        List;
    void                   *UserBuffer;
    __u32                   UserBufferSize;
    void                   *UserCompletionRoutine;
    void                   *UserData;

} ASYNC_URB,*PASYNC_URB;

typedef struct __WRG3
{
    char Name[32];

    int DeviceIndex;
    int Minor;

    struct device ClassDevice;

    struct cdev cdev;
    WRG3_MUTEX Ioctl_Lock;


    __u8 InterfaceType;
    __u8 Opened;
    __u8 Connected;

    union
    {
        struct //PCI device data
        {
            struct pci_dev             *PCIDevice;
            __u8                       *MemHPI;
            __u8                       *MemGP;

            int                         IrqNum;
            __u8                        IrqHandlerRegistered;
            struct semaphore            Interrupt_Semaphore;
            __u8                        InterruptStarted;
            spinlock_t                  Interrupt_SpinLock;
            __u32                       InterruptRequest;
            __u16                      *Buffer;
        } PCI;


        struct //USB device data
        {
            struct usb_device          *USBDevice;
            struct usb_interface       *USBInterface;
            __u8                        NumberOfEndPoints;

            volatile int                AsyncUrbs_Active;
            struct list_head            AsyncUrbs_Submitted;
            struct list_head            AsyncUrbs_Complete;
            spinlock_t                  AsyncUrbs_SpinLock;
            unsigned int                AsyncUrbs_SeqNum;
            struct semaphore            AsyncUrbs_Semaphore;
            struct completion           AsyncUrbs_Completion;
        } USB;
    };
} WRG3,*PWRG3;

static PWRG3 WRG3_Devices[MAX_WRG3_DEVICES];

#ifdef USE_MUTEX
static DEFINE_MUTEX(WRG3_Device_Lock);
#else
static DECLARE_MUTEX(WRG3_Device_Lock);
#endif


static struct usb_device_id WRG3USB_ID_Table[]=
{
    {USB_DEVICE(0x14E0,0x3105)},
    {},
};

static struct pci_device_id WRG3PCI_ID_Table[]=
{
    {.vendor=0x104C,.device=0xAC60,.subvendor=0x186F,.subdevice=0x3105},
    {}, 
};

MODULE_DEVICE_TABLE(usb,WRG3USB_ID_Table);
MODULE_DEVICE_TABLE(pci,WRG3PCI_ID_Table);


int RegisterWRG3Device(PWRG3 Dev);
int UnregisterWRG3Device(PWRG3 Dev);
int WRG3PCI_StopInterrupt(PWRG3 Dev);


int WRG3_Release(struct inode *inode, struct file *filp);
int WRG3_Open(struct inode *inode, struct file *filp);

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,37)
long WRG3_Ioctl( struct file *filp, unsigned int cmd, unsigned long arg);
#else
int WRG3_Ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);
#endif

long WRG3_Compat_Ioctl(struct file *filp, unsigned int cmd, unsigned long arg);

static struct file_operations WRG3_File_Operations=
{
    .open=WRG3_Open,
    .release=WRG3_Release,
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,37)
    .unlocked_ioctl=WRG3_Ioctl,
#else
    .ioctl=WRG3_Ioctl,
#endif
    
#ifdef CONFIG_COMPAT
    .compat_ioctl   = WRG3_Compat_Ioctl
#endif
};

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,26)

static void WRG3ReleaseCD(struct device *cd);

static struct class WRG3_Class=
{
    .name=WRG3_NAME,
    .dev_release=WRG3ReleaseCD,
};

#else
static void WRG3ReleaseCD(struct class_device *cd);

static struct class WRG3_Class=
{
    .name=WRG3_NAME,
    .release=WRG3ReleaseCD,
};

#endif

void WRG3PCI_FreeResources(PWRG3 Dev)
{
    if(Dev->PCI.IrqHandlerRegistered)
    {
        free_irq(Dev->PCI.IrqNum,Dev);
        Dev->PCI.IrqHandlerRegistered=0;
    }

    if(Dev->PCI.MemHPI!=NULL)
    {
        iounmap(Dev->PCI.MemHPI);
        Dev->PCI.MemHPI=NULL;
    }

    if(Dev->PCI.MemGP!=NULL)
    {
        iounmap(Dev->PCI.MemGP);
        Dev->PCI.MemGP=NULL;
    }
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
irqreturn_t WRG3PCI_InterruptHandler(int irqno, void *dev_id)
#else
irqreturn_t WRG3PCI_InterruptHandler(int irqno, void *dev_id, struct pt_regs *regs)
#endif
{
 PWRG3 Dev;
 __u32 Request;
 unsigned long Flags;
 
    Dev=(PWRG3)dev_id;
 
    Request=ioread32(Dev->PCI.MemHPI+0x04);

    if(Request==0 || Request==0xFFFFFFFF)
    {
        return IRQ_NONE;
    }


    spin_lock_irqsave(&Dev->PCI.Interrupt_SpinLock,Flags);

    if(Dev->PCI.InterruptStarted!=0)
    {
        if(Dev->PCI.InterruptRequest==0)
        {
            up(&Dev->PCI.Interrupt_Semaphore);
        }
        Dev->PCI.InterruptRequest|=Request;
    }

    spin_unlock_irqrestore(&Dev->PCI.Interrupt_SpinLock,Flags);

    iowrite32(Request,Dev->PCI.MemHPI+0x04);

    return IRQ_HANDLED;
}

int WRG3PCI_ReadPIC(PWRG3 Dev,__u8 Address,__u8 *Value)
{
 __u16 RegistryAddress;
 int Retry;
 __u8 *Status;

    RegistryAddress=(Address & 0x3F) | 0x80;

    Status=Dev->PCI.MemGP+4;

    Retry=MAX_PIC_RETRY;


    while(Retry--)
    {
        if(!(ioread16(Status)&0x0100))
        {
            break;
        }
    }

    if(Retry<=0)
    {
        ioread16(Dev->PCI.MemGP);

        Retry=MAX_PIC_RETRY;

        while(Retry--)
        {
            if(!(ioread16(Status)&0x0100))
            {
                break;
            }
        }

        if(Retry<=0)
        {
            return -EIO;
        }
    }

    iowrite16(RegistryAddress,Dev->PCI.MemGP);

    Retry=MAX_PIC_RETRY;

    while(Retry--)
    {
        if(ioread16(Status)&0x0100)
        {
            break;
        }
    }

    if(Retry<=0)
    {
        return -EIO;
    }

    *Value=(__u8)(ioread16(Dev->PCI.MemGP) & 0xFF);


    return 0;
}

int WRG3PCI_WritePIC(PWRG3 Dev,__u8 Address,__u8 Value)
{
 __u16 RegistryAddress;
 __u8 *Status;
 int Retry;

    RegistryAddress=(Address&0x3F)|0x80|((Value&0x80)>>1);
    Status=Dev->PCI.MemGP+4;

    Retry=MAX_PIC_RETRY;


    while(Retry--)
    {
        if(!(ioread16(Status)&0x0100))
        {
            break;
        }
    }

    if(Retry<=0)
    {
        ioread16(Dev->PCI.MemGP);

        Retry=MAX_PIC_RETRY;

        while(Retry--)
        {
            if(!(ioread16(Status)&0x0100))
            {
                break;
            }
        }

        if(Retry<=0)
        {
            return -EIO;
        }
    }

    iowrite16(RegistryAddress,Dev->PCI.MemGP);

    Retry=MAX_PIC_RETRY;

    while(Retry--)
    {
        if((ioread16(Status)&0x0100))
        {
            break;
        }
    }

    if(Retry<=0)
    {
        return -EIO;
    }

    iowrite16(Value&0x7F,Dev->PCI.MemGP);

    return 0;
}

static int WRG3PCI_Probe(struct pci_dev *dev, const struct pci_device_id *id)
{
 PWRG3 Dev;
 int Error;
 unsigned long ResourceStart;
 unsigned long ResourceLen;
 int i;
 int MemResIndex;

    Dev=kmalloc(sizeof(WRG3)+PCI_BUFFER_SIZE,GFP_KERNEL);

    if(Dev==NULL)
    {
        return -ENOMEM;
    }

    Error=pci_enable_device(dev);

    if(Error)
    {
        kfree(Dev);
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to enable pci device (%d)\n",Error);
        return Error;
    }

    memset(Dev,0,sizeof(WRG3));

    Dev->PCI.PCIDevice=dev;
    Dev->InterfaceType=WRG3_PCI;
    Dev->DeviceIndex=-1;
    Dev->PCI.Buffer=(__u16*)(Dev+1);
    strcpy(Dev->Name,WRG3_DISPLAY_NAME);

    Dev->PCI.InterruptStarted=0;

    sema_init(&Dev->PCI.Interrupt_Semaphore,0);

    spin_lock_init(&Dev->PCI.Interrupt_SpinLock);

    WRG3_INIT_MUTEX(Dev->Ioctl_Lock);

    WRG3_LOCK(Dev->Ioctl_Lock);

    for(i=0,MemResIndex=0,Error=0;i<DEVICE_COUNT_RESOURCE && MemResIndex<3 && Error==0;i++)
    {
        if(pci_resource_flags(dev,i) & IORESOURCE_MEM)
        {
            ResourceStart=pci_resource_start(dev,i);
            ResourceLen=pci_resource_len(dev,i);

            switch(MemResIndex)
            {
                case 0:
#if LINUX_VERSION_CODE>=KERNEL_VERSION(5,6,0)
                    Dev->PCI.MemHPI=(__u8*)ioremap(ResourceStart,ResourceLen);
#else
                    Dev->PCI.MemHPI=(__u8*)ioremap_nocache(ResourceStart,ResourceLen);
#endif
                    if(Dev->PCI.MemHPI==NULL)
                    {
                        Error=-ENOMEM;
                    }
                    break;

                case 2:
#if LINUX_VERSION_CODE>=KERNEL_VERSION(5,6,0)
                    Dev->PCI.MemGP=(__u8*)ioremap(ResourceStart,ResourceLen);
#else
                    Dev->PCI.MemGP=(__u8*)ioremap_nocache(ResourceStart,ResourceLen);
#endif
                    if(Dev->PCI.MemGP==NULL)
                    {
                        Error=-ENOMEM;
                    }
                    break;
                default:
                    break;
            }

            MemResIndex++;
        }
    }

    if(Dev->PCI.MemHPI==NULL || Dev->PCI.MemGP==NULL)
    {
        printk(KERN_ERR ": Failed to map memory resources.\n");
        Error=-ENOMEM;
    }

    if(Error!=0)
    {
        WRG3PCI_FreeResources(Dev);

        WRG3_UNLOCK(Dev->Ioctl_Lock);
        WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);

        kfree(Dev);
        return Error;
    }

    iowrite32(0xffffffff,Dev->PCI.MemHPI+0x04);
    iowrite32(0xffffffff,Dev->PCI.MemHPI+0x0C); 



    Dev->PCI.IrqNum=dev->irq;

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,22)
    Error=request_irq(dev->irq,WRG3PCI_InterruptHandler,IRQF_SHARED,WRG3_NAME"pci",Dev);
#else
    Error=request_irq(dev->irq,WRG3PCI_InterruptHandler,SA_INTERRUPT | SA_SHIRQ,WRG3_NAME"pci",Dev);
#endif

    if(Error!=0)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to register interrupt handler (%d)\n",Error);
        WRG3PCI_FreeResources(Dev);

        WRG3_UNLOCK(Dev->Ioctl_Lock);
        WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);
        kfree(Dev);
        return Error;
    }

    Dev->PCI.IrqHandlerRegistered=1;

    pci_set_drvdata(dev,Dev);

    pci_write_config_byte(dev, 0x44, 0x34);
    pci_write_config_byte(dev, 0x45, 0x00);
    pci_write_config_byte(dev, 0x46, 0x08);
    pci_write_config_byte(dev, 0x47, 0x00);
    pci_write_config_byte(dev, 0x47, 0x08);


    Error=RegisterWRG3Device(Dev);

    if(Error!=0)
    {
        WRG3PCI_FreeResources(Dev);
        pci_set_drvdata(dev,NULL);

        WRG3_UNLOCK(Dev->Ioctl_Lock);
        WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);
        kfree(Dev);
        return Error;
    }

    WRG3PCI_WritePIC(Dev,0x05,0x00);


    Dev->Connected=1;

    WRG3_UNLOCK(Dev->Ioctl_Lock);

    return Error;
}

static void WRG3PCI_Remove(struct pci_dev *dev)
{
 PWRG3 Dev;

    Dev=(PWRG3)pci_get_drvdata(dev);

    if(Dev)
    {
        WRG3PCI_StopInterrupt(Dev);

        UnregisterWRG3Device(Dev);

        WRG3PCI_WritePIC(Dev,0x05,0x00);

        WRG3PCI_FreeResources(Dev);

        pci_disable_device(Dev->PCI.PCIDevice);
        pci_set_drvdata(Dev->PCI.PCIDevice,NULL);

        kfree(Dev);
    }
}

int WRG3USB_Configure(PWRG3 Dev,__u8 Alternate,__u8 ResetConfig);


static int WRG3USB_Probe(struct usb_interface *interface, const struct usb_device_id *id)
{
 struct usb_device *USBDevice;
 PWRG3 Dev;
 int Error; 

    if(interface->cur_altsetting->desc.bInterfaceNumber!= 0)
    {
        return -ENODEV;
    }

    USBDevice=interface_to_usbdev(interface);


    Dev=kmalloc(sizeof(WRG3),GFP_KERNEL);
	
    if(Dev==NULL)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Out of memory.\n");
        return -ENOMEM;
    }

    memset(Dev,0,sizeof(WRG3));

    Dev->USB.USBDevice=usb_get_dev(USBDevice);
    Dev->USB.USBInterface=interface;
    Dev->InterfaceType=WRG3_USB;
    Dev->DeviceIndex=-1;
    strcpy(Dev->Name,WRG3_DISPLAY_NAME);

    WRG3_INIT_MUTEX(Dev->Ioctl_Lock);


    INIT_LIST_HEAD(&Dev->USB.AsyncUrbs_Submitted);
    INIT_LIST_HEAD(&Dev->USB.AsyncUrbs_Complete);

    spin_lock_init(&Dev->USB.AsyncUrbs_SpinLock);
    Dev->USB.AsyncUrbs_Active=0;
    sema_init(&Dev->USB.AsyncUrbs_Semaphore,0);


    usb_set_intfdata(interface,Dev);


    WRG3_LOCK(Dev->Ioctl_Lock);

    Error=RegisterWRG3Device(Dev);

    if(Error!=0)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        goto Error_Exit;
    }
    else
    {
        usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),0x09,USB_TYPE_VENDOR,0,0,NULL,0,TO_JIFFIES(USB_TIMEOUT));
        usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),0x03,USB_TYPE_VENDOR,0xA8,0,NULL,0,TO_JIFFIES(USB_TIMEOUT));
    }

    Dev->Connected=1;

    WRG3_UNLOCK(Dev->Ioctl_Lock);

    return 0;

    Error_Exit:

    usb_set_intfdata(interface,NULL);
    usb_put_dev(Dev->USB.USBDevice);

    WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);

    kfree(Dev);

    return Error;
}


void WRG3USB_Free(PWRG3 Dev)
{

}

void WRG3USB_KillAsyncUrbs(PWRG3 Dev)
{
 unsigned long Flags;
 struct urb *Urb;

    spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

    while(!list_empty(&Dev->USB.AsyncUrbs_Submitted))
    {
        Urb=list_entry(Dev->USB.AsyncUrbs_Submitted.prev,ASYNC_URB,List)->Urb;
        usb_get_urb(Urb);
        spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        usb_kill_urb(Urb);
        usb_free_urb(Urb);

        spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);
    }

    spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);
}

void WRG3USB_WaitForAsyncUrbs(PWRG3 Dev)
{
 int i;
 unsigned long Flags;

    spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

    for(i=0;!list_empty(&Dev->USB.AsyncUrbs_Submitted) && i<200;i++)
    {
        spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        msleep(10);

        spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);
    }

    spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);
}

int WRG3USB_StopAsyncMessages(PWRG3 Dev);

static void WRG3USB_Disconnect(struct usb_interface *interface)
{
 PWRG3 Dev;

    Dev=usb_get_intfdata(interface);

    if(Dev==NULL)
    {
        return;
    }

    WRG3_LOCK(Dev->Ioctl_Lock);

    WRG3USB_StopAsyncMessages(Dev);

    WRG3USB_WaitForAsyncUrbs(Dev);

    //WRG3USB_Configure(Dev,0,0);

    UnregisterWRG3Device(Dev);

    //usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),0x09,USB_TYPE_VENDOR,0,0,NULL,0,TO_JIFFIES(500));
    //usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),0x03,USB_TYPE_VENDOR,0x57,0,NULL,0,TO_JIFFIES(500));        

    if(!Dev->Opened)
    {
        usb_set_intfdata(interface,NULL);

        usb_put_dev(Dev->USB.USBDevice);

        WRG3_UNLOCK(Dev->Ioctl_Lock);

        WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);

        kfree(Dev);
    }
    else
    {
        Dev->Connected=0;

        WRG3_UNLOCK(Dev->Ioctl_Lock);
    }
}



static struct usb_driver WRG3USB_Driver= 
{
    .name=WRG3_NAME,
    .probe=WRG3USB_Probe,
    .disconnect=WRG3USB_Disconnect,
    .id_table=WRG3USB_ID_Table,
};

static struct pci_driver WRG3PCI_Driver = 
{
    .name=WRG3_NAME,
    .id_table=WRG3PCI_ID_Table,
    .probe=WRG3PCI_Probe,
    .remove=WRG3PCI_Remove,
};

static int __init WRG3_Init(void)
{
 int Error;
 dev_t FirstDev;

    FirstDev=MKDEV(WiNRADiO_MAJOR,WRG3_MINOR_BASE);

    Error=register_chrdev_region(FirstDev,MAX_WRG3_DEVICES,WRG3_NAME);

    if(Error<0)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to register chrdev region (%d)\n",Error);
        return Error;
    }

    Error=class_register(&WRG3_Class);

    if(Error<0)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to register class (%d)\n",Error);
        unregister_chrdev_region(FirstDev,MAX_WRG3_DEVICES);
        return Error;
    }

    Error=usb_register(&WRG3USB_Driver);

    if(Error!=0)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to register USB driver (%d)\n",Error);
    }	
    else
    {
        Error=pci_register_driver(&WRG3PCI_Driver);

        if(Error!=0)
        {
            printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to register PCI driver (%d)\n",Error);
        }
    }

    if(Error!=0)
    {
        class_unregister(&WRG3_Class);
        unregister_chrdev_region(FirstDev,MAX_WRG3_DEVICES);
    }

    return Error;
}

static void __exit WRG3_Exit(void)
{
 dev_t FirstDev;

    usb_deregister(&WRG3USB_Driver);
    pci_unregister_driver(&WRG3PCI_Driver);

    class_unregister(&WRG3_Class);

    FirstDev=MKDEV(WiNRADiO_MAJOR,WRG3_MINOR_BASE);
    unregister_chrdev_region(FirstDev,MAX_WRG3_DEVICES); 
}



int WRG3USB_Open(PWRG3 Dev)
{
    WRG3USB_Configure(Dev,0,1);
    usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),0x03,USB_TYPE_VENDOR,0xA0,0,NULL,0,TO_JIFFIES(USB_TIMEOUT));
    return 0;
}

int WRG3PCI_Open(PWRG3 Dev)
{
    return 0;
}

int WRG3_Open(struct inode *inode, struct file *filp)
{
 PWRG3 Dev;
 int Error;

    Dev=container_of(inode->i_cdev,WRG3,cdev);

    if(Dev==NULL)
    {
        return -ENOTTY;
    }

    if(WRG3_LOCK_INTERRUPTIBLE(Dev->Ioctl_Lock))
    {
        return -ERESTARTSYS;
    }

    if(Dev->Opened)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EBUSY;
    }

    filp->private_data=Dev;

    switch(Dev->InterfaceType)
    {
        case WRG3_USB:
            Error=WRG3USB_Open(Dev);
            break;
        case WRG3_PCI:
            Error=WRG3PCI_Open(Dev);
            break;
        default:
            Error=-ENOTTY;
            break;
    }

    if(Error==0)
    {
        Dev->Opened=1;
    }

    WRG3_UNLOCK(Dev->Ioctl_Lock);

    return Error;
}

int WRG3PCI_Reset(PWRG3 Dev);

int WRG3_Release(struct inode *inode, struct file *filp)
{
 PWRG3 Dev;

    Dev=(PWRG3)filp->private_data;

    if(Dev!=NULL)
    { 
        WRG3_LOCK(Dev->Ioctl_Lock);

        Dev->Opened=0;

        if(Dev->InterfaceType==WRG3_USB)
        {
            if(Dev->Connected)
            {
                usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice, 0),0x2C,USB_TYPE_VENDOR,5,0,NULL,0,USB_CTRL_SET_TIMEOUT);
                usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),0x09,USB_TYPE_VENDOR,0,0,NULL,0,TO_JIFFIES(USB_TIMEOUT));
                usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),0x03,USB_TYPE_VENDOR,0xA8,0,NULL,0,TO_JIFFIES(USB_TIMEOUT));

                WRG3USB_StopAsyncMessages(Dev);

                WRG3USB_WaitForAsyncUrbs(Dev);

                WRG3USB_Configure(Dev,0,0);
            }
            else
            {
                usb_set_intfdata(Dev->USB.USBInterface,NULL);
                usb_put_dev(Dev->USB.USBDevice);


                WRG3_UNLOCK(Dev->Ioctl_Lock);

                WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);

                kfree(Dev);
            }
        }
        else
        {
            WRG3PCI_StopInterrupt(Dev);
            WRG3PCI_WritePIC(Dev,0x05,0x00);
            WRG3PCI_Reset(Dev);
        }

        WRG3_UNLOCK(Dev->Ioctl_Lock);
    }

    return 0;
}


int WRG3USB_ControlMessage_Send(PWRG3 Dev,unsigned long arg)
{
 int Error;
 void *Buff;
 USB_CONTROL_MSG UsbMsg;
 
    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -ENOTTY;
    }

    if(copy_from_user(&UsbMsg,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    if(UsbMsg.Buffer!=NULL && UsbMsg.Size!=0)
    {
        Buff=kmalloc(UsbMsg.Size,GFP_KERNEL);

        if(Buff==NULL)
        {
            return -ENOMEM;
        }

        if(copy_from_user(Buff,UsbMsg.Buffer,UsbMsg.Size))
        {
            kfree(Buff);
            return -EFAULT;
        }
    }
    else
    {
        Buff=NULL;
        UsbMsg.Size=0;
    }

    Error=usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),
                          UsbMsg.Request,USB_TYPE_VENDOR,
                          UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0));
        usb_clear_halt(Dev->USB.USBDevice,usb_rcvctrlpipe(Dev->USB.USBDevice,0));

        Error=usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),
                              UsbMsg.Request,USB_TYPE_VENDOR,
                              UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));
    }

    if(Error>=0)
    {
        if(Error==UsbMsg.Size)
        {
            Error=0;
        }
        else
        {
            Error=-EIO;
        }
    }

    if(Buff!=NULL)
    {
        kfree(Buff);
    }

    return Error;
}

int WRG3USB_ControlMessage_Receive(PWRG3 Dev,unsigned long arg)
{
 USB_CONTROL_MSG UsbMsg;
 void *Buff;
 int Error;

 
    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -ENOTTY;
    }

    if(copy_from_user(&UsbMsg,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    if(UsbMsg.Buffer!=NULL && UsbMsg.Size!=0)
    {
        Buff=kmalloc(UsbMsg.Size,GFP_KERNEL);

        if(Buff==NULL)
        {
            return -ENOMEM;
        }
    }
    else
    {
        Buff=NULL;
        UsbMsg.Size=0;
    }

    Error=usb_control_msg(Dev->USB.USBDevice,usb_rcvctrlpipe(Dev->USB.USBDevice,0),
                          UsbMsg.Request,USB_DIR_IN | USB_TYPE_VENDOR,
                          UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));

    if(Buff!=NULL)
    {
        if(Error>0)
        {
            if(copy_to_user((void __user*)UsbMsg.Buffer,Buff,Error))//UsbMsg.Size))
            {
                Error=-EFAULT;
            }
        }

        kfree(Buff);
    }

    return Error;
}

int WRG3USB_ControlMessage(PWRG3 Dev,unsigned long arg)
{
 USB_CONTROL_MSG UsbMsg;
 int Error;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -ENOTTY;
    }

    if(copy_from_user(&UsbMsg,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    Error=usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),
                UsbMsg.Request,USB_TYPE_VENDOR,
                UsbMsg.Value,UsbMsg.Index,NULL,0,TO_JIFFIES(UsbMsg.Timeout));

    if(Error>=0)
    {
        Error=0;
    }

    return Error;
}

int WRG3PCI_ReadPICBytes(PWRG3 Dev,unsigned long arg)
{
 int Error;
 PIC_REG Reg;
 __u16 i;
 __u8 *Addresses;
 __u8 *Values;
 __u8 AddressesBuff[256];
 __u8 ValuesBuff[256];
 __u8 *Address;
 __u8 *Value;
 
    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    } 
 
    Error=0;
 

    if(copy_from_user(&Reg,(void __user*)arg,sizeof(Reg)))
    {
        return -EFAULT;
    }

    if(Reg.ByteCount==0)
    {
        return 0;
    }

    if(Reg.ByteCount>sizeof(AddressesBuff))
    {
        Addresses=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Addresses==NULL)
        {
            return -ENOMEM;
        }

        Values=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Values==NULL)
        {
            kfree(Addresses);
            return -ENOMEM;
        }
    }
    else
    {
        Addresses=AddressesBuff;
        Values=ValuesBuff;
    }

    if(copy_from_user(Addresses,(void __user*)Reg.Addresses,Reg.ByteCount)==0)
    {
        Address=Addresses;
        Value=Values;

        for(i=0;i<Reg.ByteCount;i++,Address++,Value++)
        {
            Error=WRG3PCI_ReadPIC(Dev,*Address,Value);

            if(Error!=0)
            {
                break;
            }
        }

        if(Error==0)
        {
            if(copy_to_user((void __user*)Reg.Values,Values,Reg.ByteCount))
            {
                Error=-EFAULT;
            }
        }
    }
    else
    {
        Error=-EFAULT;
    }

    if(Addresses!=AddressesBuff)
    {
        kfree(Addresses);
        kfree(Values);
    }

    return Error;
}

int WRG3PCI_WritePICBytes(PWRG3 Dev,unsigned long arg)
{
 int Error;
 PIC_REG Reg;
 __u16 i;
 __u8 *Addresses;
 __u8 *Values;
 __u8 AddressesBuff[256];
 __u8 ValuesBuff[256];
 __u8 *Address;
 __u8 *Value;

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    } 

    Error=0;

    if(copy_from_user(&Reg,(void __user*)arg,sizeof(Reg)))
    {
        return -EFAULT;
    }

    if(Reg.ByteCount==0)
    {
        return 0;
    }


    if(Reg.ByteCount>sizeof(AddressesBuff))
    {
        Addresses=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Addresses==NULL)
        {
            return -ENOMEM;
        }

        Values=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Values==NULL)
        {
            kfree(Addresses);
            return -ENOMEM;
        }
    }
    else
    {
        Addresses=AddressesBuff;
        Values=ValuesBuff;
    }

    if(copy_from_user(Addresses,(void __user*)Reg.Addresses,Reg.ByteCount)==0 &&
       copy_from_user(Values,(void __user*)Reg.Values,Reg.ByteCount)==0)
    {
        Address=Addresses;
        Value=Values;

        for(i=0;i<Reg.ByteCount;i++,Address++,Value++)
        {
            Error=WRG3PCI_WritePIC(Dev,*Address,*Value);

            if(Error!=0)
            {
                break;
            }
        }
    }
    else
    {
        Error=-EFAULT;
    }

    if(Addresses!=AddressesBuff)
    {
        kfree(Addresses);
        kfree(Values);
    }

    return Error;
}

int WRG3PCI_WritePICByte(PWRG3 Dev,unsigned long arg)
{
 __u8 PicReg[2];
 
    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    } 

    if(get_user((*((__u16*)PicReg)),(__u16 __user*)arg))
    {
        return -EFAULT;
    }

    return WRG3PCI_WritePIC(Dev,PicReg[0],PicReg[1]);
}

int WRG3PCI_ReadPICByte(PWRG3 Dev,unsigned long arg)
{
 __u8 PicReg[2];
 int Error;
 
    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    } 
 

    if(get_user(*((__u16*)PicReg),(__u16 __user*)arg))
    {
        return -EFAULT;
    }

    Error=WRG3PCI_ReadPIC(Dev,PicReg[0],PicReg+1);

    if(Error==0)
    {
        if(put_user(*((__u16*)PicReg),(__u16 __user*)arg))
        {
            Error=-EFAULT;
        }
    }

    return Error;
}




#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,19)
static void WRG3USB_AsyncUrbComplete(struct urb *Urb)
#else
static void WRG3USB_AsyncUrbComplete(struct urb *Urb,struct pt_regs *Regs)
#endif
{
 PWRG3 Dev;
 PASYNC_URB AsyncUrb;
 unsigned long Flags;
 __u8 Free;

    AsyncUrb=(PASYNC_URB)Urb->context;
    Dev=(PWRG3)AsyncUrb->Dev;

    spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

    list_del(&AsyncUrb->List);

    if(Dev->USB.AsyncUrbs_Active && AsyncUrb->SeqNum==Dev->USB.AsyncUrbs_SeqNum)
    {
        list_add_tail(&AsyncUrb->List,&Dev->USB.AsyncUrbs_Complete);
        Free=0;
    }
    else
    {
        Free=1;
    }

    spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

    if(Free)
    {
        kfree(AsyncUrb->Urb->transfer_buffer);
        usb_free_urb(AsyncUrb->Urb);
        kfree(AsyncUrb);
    }
    else
    {
        up(&Dev->USB.AsyncUrbs_Semaphore);
    }
}


int WRG3USB_StartAsyncMessages(PWRG3 Dev)
{
    if(!Dev->USB.AsyncUrbs_Active)
    {
        Dev->USB.AsyncUrbs_Active=1;
    }

    return 0;
}

void WRG3USB_FreeAsyncUrbs(PWRG3 Dev)
{
 unsigned long Flags;
 struct list_head *ListEntry;
 PASYNC_URB AsyncUrb;


    for(;;)
    {
        spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        if(list_empty(&Dev->USB.AsyncUrbs_Complete))
        {
            spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);
            break;
        }

        ListEntry=Dev->USB.AsyncUrbs_Complete.next;

        AsyncUrb=list_entry(ListEntry,ASYNC_URB,List);

        list_del(ListEntry);

        spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        if(AsyncUrb->Urb->transfer_buffer!=NULL)
        {
            kfree(AsyncUrb->Urb->transfer_buffer);
        }

        usb_free_urb(AsyncUrb->Urb);

        kfree(AsyncUrb);
    }
}

int WRG3USB_StopAsyncMessages(PWRG3 Dev)
{
 unsigned long Flags;
 int i;

    WRG3USB_KillAsyncUrbs(Dev);

    if(Dev->USB.AsyncUrbs_Active)
    {
        spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        Dev->USB.AsyncUrbs_Active=0;
        Dev->USB.AsyncUrbs_SeqNum++;

        spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);


        for(i=0;i<10;i++)
        {
            up(&Dev->USB.AsyncUrbs_Semaphore);
        }

        WRG3USB_FreeAsyncUrbs(Dev);
    }

    return 0;
}

int WRG3USB_GetCompleteAsyncUrb(PWRG3 Dev,unsigned long arg)
{
 USB_COMPLETE_ASYNC_MESSAGE Result;
 struct list_head *ListEntry;
 PASYNC_URB AsyncUrb;
 unsigned long Flags;
 __u8 *UserBuffer,*Buffer;
 int i;
 struct urb *Urb;

    if(!Dev->Connected || !Dev->USB.AsyncUrbs_Active)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EINVAL;
    }


    spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

    while(list_empty(&Dev->USB.AsyncUrbs_Complete))
    {
        spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        WRG3_UNLOCK(Dev->Ioctl_Lock);

        if(down_interruptible(&Dev->USB.AsyncUrbs_Semaphore))
        {
            return -ERESTARTSYS;
        }

        if(!Dev->Connected)
        {
            return -EINVAL;
        }

        if(!Dev->USB.AsyncUrbs_Active)
        {
            return -EINVAL;
        }

        if(WRG3_LOCK_INTERRUPTIBLE(Dev->Ioctl_Lock))
        {
            return -ERESTARTSYS;
        }


        spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);
    }

    ListEntry=Dev->USB.AsyncUrbs_Complete.next;
    AsyncUrb=list_entry(ListEntry,ASYNC_URB,List);
    list_del(ListEntry);

    spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);


    Urb=AsyncUrb->Urb;

    Result.CompletionRoutine=AsyncUrb->UserCompletionRoutine;
    Result.Buffer=AsyncUrb->UserBuffer;
    Result.BufferSize=AsyncUrb->UserBufferSize;
    Result.Transferred=Urb->actual_length;
    Result.Status=Urb->status;
    Result.Data=AsyncUrb->UserData;

    if(Urb->transfer_buffer!=NULL)
    {
        if(Result.Buffer!=NULL && Result.Status==0 && usb_pipein(Urb->pipe))
        {
            if(usb_pipeisoc(Urb->pipe))
            {
                Result.Transferred=0;
                UserBuffer=(__u8*)Result.Buffer;
                Buffer=(__u8*)Urb->transfer_buffer;

                for(i=0;i<Urb->number_of_packets;i++)
                {
                    if(Urb->iso_frame_desc[i].status==0 && Urb->iso_frame_desc[i].actual_length>0)
                    {
                        if(copy_to_user((void __user*)UserBuffer,Buffer+Urb->iso_frame_desc[i].offset,Urb->iso_frame_desc[i].actual_length))
                        {
                            Result.Status=-EFAULT;
                            Result.Transferred=0;
                            break;
                        }
                        UserBuffer+=Urb->iso_frame_desc[i].actual_length;
                        Result.Transferred+=Urb->iso_frame_desc[i].actual_length;
                    }
                }
            }
            else
            {
                if(copy_to_user((void __user*)Result.Buffer,Urb->transfer_buffer,Urb->actual_length))
                {
                    Result.Status=-EFAULT;
                    Result.Transferred=0;
                }
            }
        }

        kfree(Urb->transfer_buffer);
    }

    usb_free_urb(Urb);

    kfree(AsyncUrb);

    if(copy_to_user((void __user*)arg,&Result,sizeof(Result)))
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EFAULT;
    }

    WRG3_UNLOCK(Dev->Ioctl_Lock);
    return 0;
}


int WRG3USB_WaitForIdle(PWRG3 Dev)
{
 uint16_t Status[4];
 uint8_t RetryCount;

    RetryCount=100;

    while(RetryCount--)
    {

        if(usb_control_msg(Dev->USB.USBDevice,usb_rcvctrlpipe(Dev->USB.USBDevice,0),
                          0x0A,USB_DIR_IN | USB_TYPE_VENDOR,
                          0,0,Status,sizeof(Status),TO_JIFFIES(2000)))
        {
            return -1;
        }

        if((Status[0]&0x0002)==0)
        {
            return 0;
        }

        msleep(10);
    }
    return -1;
}

int WRG3USB_Configure(PWRG3 Dev,__u8 Alternate,__u8 ResetConfig)
{
 __u32 i;
 int Error;

    WRG3USB_KillAsyncUrbs(Dev);
    WRG3USB_WaitForAsyncUrbs(Dev);

    WRG3USB_WaitForIdle(Dev);


    if(ResetConfig)
    {
        Error=usb_reset_configuration(Dev->USB.USBDevice);

        if(Error!=0)
        {
            printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to reset configuration (%d).\n",Error);
            return Error;
        }
    }

    Error=usb_set_interface(Dev->USB.USBDevice,0,Alternate);//Dev->USB.USBInterface->altsetting[0].desc.bInterfaceNumber

    if(Error!=0)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to set alternative setting %u for interface (%d).\n",Alternate,Error);
        return Error;
    }

    Dev->USB.NumberOfEndPoints=Dev->USB.USBInterface->cur_altsetting->desc.bNumEndpoints;

    for(i=0;i<MAX_ENDPOINTS;i++)
    {
        if(i>0)
        {
            if(Dev->USB.USBDevice->ep_in[i]!=NULL)
            {
                Error=usb_control_msg(Dev->USB.USBDevice, usb_sndctrlpipe(Dev->USB.USBDevice, 0),
                                      USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT,
                                      USB_ENDPOINT_HALT, i|USB_DIR_IN, NULL, 0,
                                      USB_CTRL_SET_TIMEOUT);
                
            }

            if(Dev->USB.USBDevice->ep_out[i]!=NULL)
            {
                Error=usb_control_msg(Dev->USB.USBDevice, usb_sndctrlpipe(Dev->USB.USBDevice, 0),
                                      USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT,
                                      USB_ENDPOINT_HALT, i, NULL, 0,
                                      USB_CTRL_SET_TIMEOUT);

            }
        }
    }

    return 0;
}

int WRG3USB_GetEndPointInfo(PWRG3 Dev,unsigned long arg)
{
 ENDPOINTS EndPoints;
 __u32 i;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    EndPoints.Count=Dev->USB.NumberOfEndPoints;

    for(i=0;i<EndPoints.Count;i++)
    {
        EndPoints.EndPoint[i].Address=Dev->USB.USBInterface->cur_altsetting->endpoint[i].desc.bEndpointAddress;
        EndPoints.EndPoint[i].Number=EndPoints.EndPoint[i].Address&USB_ENDPOINT_NUMBER_MASK;
        EndPoints.EndPoint[i].MaxPacketSize=le16_to_cpu(Dev->USB.USBInterface->cur_altsetting->endpoint[i].desc.wMaxPacketSize);
    }

    if(copy_to_user((ENDPOINTS __user*)arg,&EndPoints,sizeof(EndPoints)))
    {
        return -EFAULT;
    }

    return 0;
}


int WRG3USB_BulkMessage_Send_Async2(PWRG3 Dev,USB_ASYNC_MESSAGE *Msg)//unsigned long arg)
{
 __u8 *Buffer;
 int Error;
 unsigned long Flags;
 PASYNC_URB AsyncUrb;

    if(Msg->EndPoint>15 || Dev->USB.USBDevice->ep_out[Msg->EndPoint]==NULL || Msg->Buffer==NULL || Msg->BufferSize==0)
    {
        return -EINVAL;
    }


    Buffer=(__u8*)kmalloc(Msg->BufferSize,GFP_KERNEL);

    if(Buffer==NULL)
    {
        return -ENOMEM;
    }

    if(copy_from_user(Buffer,(__u8 __user*)Msg->Buffer,Msg->BufferSize))
    {
        Error=-EFAULT;
    }
    else
    {
        AsyncUrb=(PASYNC_URB)kmalloc(sizeof(ASYNC_URB),GFP_KERNEL);

        if(AsyncUrb==NULL)
        {
            kfree(Buffer);
            return -ENOMEM;
        }

        AsyncUrb->Dev=Dev;
        AsyncUrb->UserBuffer=NULL;
        AsyncUrb->UserBufferSize=0;
        AsyncUrb->UserCompletionRoutine=Msg->CompletionRoutine;
        AsyncUrb->SeqNum=Dev->USB.AsyncUrbs_SeqNum;
        AsyncUrb->UserData=Msg->Data;

        AsyncUrb->Urb=usb_alloc_urb(0,GFP_KERNEL);

        if(AsyncUrb->Urb!=NULL)
        {
            usb_fill_bulk_urb(AsyncUrb->Urb,Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg->EndPoint),Buffer,
                              Msg->BufferSize,WRG3USB_AsyncUrbComplete,AsyncUrb);

            spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

            list_add_tail(&AsyncUrb->List,&Dev->USB.AsyncUrbs_Submitted);
            AsyncUrb->SeqNum=Dev->USB.AsyncUrbs_SeqNum;
            spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);


            Error=usb_submit_urb(AsyncUrb->Urb,GFP_KERNEL);

            if(Error==-EPIPE)
            {
                usb_clear_halt(Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg->EndPoint));

                Error=usb_submit_urb(AsyncUrb->Urb,GFP_KERNEL);
            }

            if(Error!=0)
            {
                spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

                list_del(&AsyncUrb->List);

                spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

                usb_free_urb(AsyncUrb->Urb);

                kfree(AsyncUrb);
            }
        }
        else
        {
            kfree(AsyncUrb);
            Error=-ENOMEM;
        }

        if(Error!=0)
        {
            if(Error==-EPIPE)
            {
                usb_clear_halt(Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg->EndPoint));
            }

            kfree(Buffer);
        }
    }

    return Error;
}

int WRG3USB_BulkMessage_Send_Async(PWRG3 Dev,unsigned long arg)
{
 USB_ASYNC_MESSAGE Msg;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(copy_from_user(&Msg,(USB_ASYNC_MESSAGE __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    return WRG3USB_BulkMessage_Send_Async2(Dev,&Msg);
}


int WRG3USB_BulkMessage_Send(PWRG3 Dev,unsigned long arg)
{
 USB_BULK_MESSAGE Msg;
 __u8 *Buff;
 int Error;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }
 

    if(copy_from_user(&Msg,(USB_BULK_MESSAGE __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    if(Msg.EndPoint>15 || Dev->USB.USBDevice->ep_out[Msg.EndPoint]==NULL || Msg.Buffer==NULL || Msg.BufferSize==0)
    {
        return -EINVAL;
    }


    Buff=(__u8*)kmalloc(Msg.BufferSize,GFP_KERNEL);

    if(Buff==NULL)
    {
        return -ENOMEM;
    }

    if(copy_from_user(Buff,(__u8 __user*)Msg.Buffer,Msg.BufferSize))
    {
        Error=-EFAULT;
    }
    else
    {
        Error=usb_bulk_msg(Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg.EndPoint),
                           Buff,Msg.BufferSize,&Msg.Transferred,TO_JIFFIES(Msg.Timeout));

        if(Error==-EPIPE)
        {
            usb_clear_halt(Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg.EndPoint));

            Error=usb_bulk_msg(Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg.EndPoint),
                           Buff,Msg.BufferSize,&Msg.Transferred,TO_JIFFIES(Msg.Timeout));
        }

        if(Error==0)
        {
            if(copy_to_user((USB_BULK_MESSAGE __user*)arg,&Msg,sizeof(Msg)))
            {
                Error=-EFAULT;
            }
        }
        else
        {
            if(Error==-EPIPE)
            {
                usb_clear_halt(Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg.EndPoint));
            }
        }
    }

    kfree(Buff);

    return Error;
}

int WRG3USB_IsocMessage_Receive_Async2(PWRG3 Dev,USB_ASYNC_MESSAGE *Msg)
{
 int Error;
 PASYNC_URB AsyncUrb;
 unsigned long Flags;
 struct urb *Urb;
 int NumberOfPackets;
 int Offset;
 int MaxPacketSize,i;
 __u8 *Buffer;

    if(Msg->EndPoint>15 || Dev->USB.USBDevice->ep_in[Msg->EndPoint]==NULL || Msg->Buffer==NULL || Msg->BufferSize==0)
    {
        return -EINVAL;
    }


    Buffer=(__u8*)kmalloc(Msg->BufferSize,GFP_KERNEL);

    if(Buffer==NULL)
    {
        return -ENOMEM;
    }

    AsyncUrb=(PASYNC_URB)kmalloc(sizeof(ASYNC_URB),GFP_KERNEL);

    if(AsyncUrb==NULL)
    {
        kfree(Buffer);
        return -ENOMEM;
    }

    AsyncUrb->Dev=Dev;
    AsyncUrb->UserBuffer=Msg->Buffer;
    AsyncUrb->UserBufferSize=Msg->BufferSize;
    AsyncUrb->UserCompletionRoutine=Msg->CompletionRoutine;
    AsyncUrb->UserData=Msg->Data;

    MaxPacketSize=Dev->USB.USBDevice->ep_in[Msg->EndPoint]->desc.wMaxPacketSize;

    NumberOfPackets=(Msg->BufferSize+MaxPacketSize-1)/MaxPacketSize;

    AsyncUrb->Urb=Urb=usb_alloc_urb(NumberOfPackets,GFP_KERNEL);

    if(AsyncUrb->Urb==NULL)
    {
        kfree(AsyncUrb);
        kfree(Buffer);
        return -ENOMEM;
    }

    Urb->dev=Dev->USB.USBDevice;
    Urb->pipe=usb_rcvisocpipe(Dev->USB.USBDevice,Msg->EndPoint);
    Urb->interval=1;
    Urb->transfer_flags=URB_ISO_ASAP;
    Urb->transfer_buffer=Buffer;
    Urb->transfer_buffer_length=Msg->BufferSize;
    Urb->number_of_packets=NumberOfPackets;
    Urb->complete=WRG3USB_AsyncUrbComplete;
    Urb->context=AsyncUrb;

    for(i=0,Offset=0;i<NumberOfPackets;i++,Offset+=MaxPacketSize)
    {
        Urb->iso_frame_desc[i].offset=Offset;
        Urb->iso_frame_desc[i].length=Msg->BufferSize>=MaxPacketSize?MaxPacketSize:Msg->BufferSize;
        Msg->BufferSize-=MaxPacketSize;
    }


    spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

    AsyncUrb->SeqNum=Dev->USB.AsyncUrbs_SeqNum;
    list_add_tail(&AsyncUrb->List,&Dev->USB.AsyncUrbs_Submitted);
    spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

    Error=usb_submit_urb(AsyncUrb->Urb,GFP_KERNEL);

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USB.USBDevice,usb_rcvisocpipe(Dev->USB.USBDevice,Msg->EndPoint));

        Error=usb_submit_urb(AsyncUrb->Urb,GFP_KERNEL);
    }

    if(Error!=0)
    {
        spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        list_del(&AsyncUrb->List);

        spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        usb_free_urb(AsyncUrb->Urb);

        kfree(AsyncUrb);
        kfree(Buffer);

        if(Error==-EPIPE)
        {
            usb_clear_halt(Dev->USB.USBDevice,usb_rcvisocpipe(Dev->USB.USBDevice,Msg->EndPoint));
        }
    }

    return Error;
}

int WRG3USB_IsocMessage_Receive_Async(PWRG3 Dev,unsigned long arg)
{
 USB_ASYNC_MESSAGE Msg;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(copy_from_user(&Msg,(USB_ASYNC_MESSAGE __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    return WRG3USB_IsocMessage_Receive_Async2(Dev,&Msg);
}

int WRG3USB_BulkMessage_Receive_Async2(PWRG3 Dev,USB_ASYNC_MESSAGE *Msg)
{
 __u8 *Buffer;
 int Error;
 PASYNC_URB AsyncUrb;
 unsigned long Flags;

    if(Msg->EndPoint>15 || Dev->USB.USBDevice->ep_in[Msg->EndPoint]==NULL || Msg->Buffer==NULL || Msg->BufferSize==0)
    {
        return -EINVAL;
    }

    Buffer=(__u8*)kmalloc(Msg->BufferSize,GFP_KERNEL);

    if(Buffer==NULL)
    {
        return -ENOMEM;
    }


    AsyncUrb=(PASYNC_URB)kmalloc(sizeof(ASYNC_URB),GFP_KERNEL);

    if(AsyncUrb==NULL)
    {
        kfree(Buffer);
        return -ENOMEM;
    }

    AsyncUrb->Dev=Dev;
    AsyncUrb->UserBuffer=Msg->Buffer;
    AsyncUrb->UserBufferSize=Msg->BufferSize;
    AsyncUrb->UserCompletionRoutine=Msg->CompletionRoutine;
    AsyncUrb->UserData=Msg->Data;

    AsyncUrb->Urb=usb_alloc_urb(0,GFP_KERNEL);

    if(AsyncUrb->Urb==NULL)
    {
        kfree(AsyncUrb);
        kfree(Buffer);
        return -ENOMEM;
    }

    usb_fill_bulk_urb(AsyncUrb->Urb,Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg->EndPoint),Buffer,
                      Msg->BufferSize,WRG3USB_AsyncUrbComplete,AsyncUrb);


    spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

    AsyncUrb->SeqNum=Dev->USB.AsyncUrbs_SeqNum;
    list_add_tail(&AsyncUrb->List,&Dev->USB.AsyncUrbs_Submitted);
    spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

    Error=usb_submit_urb(AsyncUrb->Urb,GFP_KERNEL);

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg->EndPoint));

        Error=usb_submit_urb(AsyncUrb->Urb,GFP_KERNEL);
    }

    if(Error!=0)
    {
        spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        list_del(&AsyncUrb->List);

        spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        usb_free_urb(AsyncUrb->Urb);
        kfree(AsyncUrb);
        kfree(Buffer);

        if(Error==-EPIPE)
        {
            usb_clear_halt(Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg->EndPoint));
        }
    }

    return Error;
}

int WRG3USB_BulkMessage_Receive_Async(PWRG3 Dev,unsigned long arg)
{
 USB_ASYNC_MESSAGE Msg;
 
    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(copy_from_user(&Msg,(USB_ASYNC_MESSAGE __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }


    return WRG3USB_BulkMessage_Receive_Async2(Dev,&Msg);
}

int WRG3USB_BulkMessage_Receive(PWRG3 Dev,unsigned long arg)
{
 USB_BULK_MESSAGE Msg;
 __u8 *Buff;
 int Error;
 
    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(copy_from_user(&Msg,(USB_BULK_MESSAGE __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    if(Msg.EndPoint>15 || Dev->USB.USBDevice->ep_in[Msg.EndPoint]==NULL || Msg.Buffer==NULL || Msg.BufferSize==0)
    {
        return -EINVAL;
    }

    Buff=(__u8*)kmalloc(Msg.BufferSize,GFP_KERNEL);

    if(Buff==NULL)
    {
        return -ENOMEM;
    }

    Error=usb_bulk_msg(Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg.EndPoint),
                       Buff,Msg.BufferSize,&Msg.Transferred,TO_JIFFIES(Msg.Timeout));

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg.EndPoint));

        Error=usb_bulk_msg(Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg.EndPoint),
                           Buff,Msg.BufferSize,&Msg.Transferred,TO_JIFFIES(Msg.Timeout));
    }

    if(Error==0)
    {
        if(copy_to_user((USB_BULK_MESSAGE __user*)arg,&Msg,sizeof(Msg)))
        {
            Error=-EFAULT;
        }

        if(copy_to_user((__u8 __user*)Msg.Buffer,Buff,Msg.BufferSize))
        {
            Error=-EFAULT;
        }
    }
    else
    {
        if(Error==-EPIPE)
        {
            usb_clear_halt(Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg.EndPoint));
        }
    }

    kfree(Buff);

    return Error;
}

int WRG3PCI_Write2(PWRG3 Dev,READ_WRITE_INFO *Info)
{
 __u16 *Buff,*Ptr;
 __u32 Size;
 __u8 *Mem;
 int Error;


    if(Info->Buffer==NULL || Info->Count==0)
    {
        return -EINVAL;
    }

    Size=Info->Count*sizeof(__u16);

    if(Size>PCI_BUFFER_SIZE)
    {
        Buff=(__u16*)kmalloc(Size,GFP_KERNEL);

        if(Buff==NULL)
        {
            return -ENOMEM;
        }
    }
    else
    {
        Buff=Dev->PCI.Buffer;
    }

    switch(Size)
    {
        case 2:
            Error=get_user((*(__u16*)Buff),(__u16 __user*)Info->Buffer);
            break;

        case 4:
            Error=get_user((*(__u32*)Buff),(__u32 __user*)Info->Buffer);
            break;

        default:
            if(copy_from_user(Buff,(__u16 __user*)Info->Buffer,Size))
            {
                Error=-EFAULT;
            }
            else
            {
                Error=0;
            }
            break;
    }

    if(Error!=0)
    {
        if(Buff!=Dev->PCI.Buffer)
        {
            kfree(Buff);
        }
        return Error;
    }

    Ptr=Buff;
    Mem=Dev->PCI.MemGP;

    iowrite32(Info->Address>>16,Mem+0x18);
    iowrite32(Info->Address&0xFFFF,Mem+0x18);

    Mem+=0x08;

    while(Info->Count--)
    {
        iowrite32(*Ptr++,Mem);
    }

    if(Buff!=Dev->PCI.Buffer)
    {
        kfree(Buff);
    }

    return 0;
}

int WRG3PCI_Write(PWRG3 Dev,unsigned long arg)
{
 READ_WRITE_INFO Info;

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    if(copy_from_user(&Info,(READ_WRITE_INFO __user*)arg,sizeof(Info)))
    {
        return -EFAULT;
    }

    return WRG3PCI_Write2(Dev,&Info);
}

int WRG3PCI_Read2(PWRG3 Dev,READ_WRITE_INFO *Info)
{
 __u16 *Buff,*Ptr;
 __u32 Size;
 __u8 *Mem;
 int Error;
 __u32 Count2;

    if(Info->Buffer==NULL || Info->Count==0)
    {
        return -EINVAL;
    }

    Size=Info->Count*sizeof(__u16);

    if(Size>PCI_BUFFER_SIZE)
    {
        Buff=(__u16*)kmalloc(Size,GFP_KERNEL);

        if(Buff==NULL)
        {
            return -ENOMEM;
        }
    }
    else
    {
        Buff=Dev->PCI.Buffer;
    }

    Mem=Dev->PCI.MemGP;

    iowrite32(Info->Address>>16,Mem+0x18);
    iowrite32(Info->Address&0xFFFF,Mem+0x18);

    Mem+=0x08;

    if(Info->Address&0x40000000)
    {
        ioread16(Mem);
    }

    Ptr=Buff;

    Count2=Info->Count>>2;

    while(Count2--)
    {
        Ptr[0]=ioread16(Mem);
        Ptr[1]=ioread16(Mem);
        Ptr[2]=ioread16(Mem);
        Ptr[3]=ioread16(Mem);
        Ptr+=4;
    }

    Info->Count&=3;

    while(Info->Count--)
    {
        *Ptr++=ioread16(Mem);
    }

    switch(Size)
    {
        case 2:
            Error=put_user(*(__u16*)Buff,(__u16 __user*)Info->Buffer);
            break;

        case 4:
            Error=put_user(*(__u32*)Buff,(__u32 __user*)Info->Buffer);
            break;

        case 8:
            Error=put_user(*(__u64*)Buff,(__u64 __user*)Info->Buffer);
            break;

        default:
            if(copy_to_user((__u16 __user*)Info->Buffer,Buff,Size))
            {
                Error=-EFAULT;
            }
            else
            {
                Error=0;
            }
            break;
    }

    if(Buff!=Dev->PCI.Buffer)
    {
        kfree(Buff);
    }

    return Error;
}

int WRG3PCI_Read(PWRG3 Dev,unsigned long arg)
{
 READ_WRITE_INFO Info;

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    if(copy_from_user(&Info,(void __user*)arg,sizeof(Info)))
    {
        return -EFAULT;
    }

    return WRG3PCI_Read2(Dev,&Info);
}

int WRG3PCI_Reset(PWRG3 Dev)
{
    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    pci_write_config_byte(Dev->PCI.PCIDevice, 0x47, 0x00);
    pci_write_config_byte(Dev->PCI.PCIDevice, 0x47, 0x08);

    return 0;
}

int WRG3PCI_WaitForInterrupt(PWRG3 Dev,unsigned long arg)
{
 unsigned long Flags;
 unsigned long Request;
 int Error;

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EINVAL;
    }

    if(Dev->PCI.InterruptStarted==0)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EINVAL;
    }

    spin_lock_irqsave(&Dev->PCI.Interrupt_SpinLock,Flags);

    Request=Dev->PCI.InterruptRequest;
    Dev->PCI.InterruptRequest=0;

    spin_unlock_irqrestore(&Dev->PCI.Interrupt_SpinLock,Flags);

    if(Request!=0)
    {
        Error=put_user(Request,(__u32 __user*)arg);

        WRG3_UNLOCK(Dev->Ioctl_Lock);

        return Error;
    }

    WRG3_UNLOCK(Dev->Ioctl_Lock);

    do
    {
        if(down_interruptible(&Dev->PCI.Interrupt_Semaphore))
        {
            return -ERESTARTSYS;
        }

        if(!Dev->PCI.InterruptStarted)
        {
            return -EINVAL;
        }

        spin_lock_irqsave(&Dev->PCI.Interrupt_SpinLock,Flags);

        Request=Dev->PCI.InterruptRequest;
        Dev->PCI.InterruptRequest=0;

        spin_unlock_irqrestore(&Dev->PCI.Interrupt_SpinLock,Flags);
    }
    while(Request==0);

    Error=put_user(Request,(__u32 __user*)arg);

    return Error;
}

int WRG3PCI_StartInterrupt(PWRG3 Dev)
{
 unsigned long Flags;

    spin_lock_irqsave(&Dev->PCI.Interrupt_SpinLock,Flags);

    if(Dev->PCI.InterruptStarted==0)
    {
        Dev->PCI.InterruptStarted=1;
        sema_init(&Dev->PCI.Interrupt_Semaphore,0);
        iowrite32(0x8C000000,Dev->PCI.MemHPI+0x08);
    }

    spin_unlock_irqrestore(&Dev->PCI.Interrupt_SpinLock,Flags);

    return 0;
}

int WRG3PCI_StopInterrupt(PWRG3 Dev)
{
 int i;
 unsigned long Flags;

    spin_lock_irqsave(&Dev->PCI.Interrupt_SpinLock,Flags);

    if(Dev->PCI.InterruptStarted!=0)
    {
        Dev->PCI.InterruptStarted=0;
        iowrite32(0xFFFFFFFF,Dev->PCI.MemHPI+0x04);
        iowrite32(0xFFFFFFFF,Dev->PCI.MemHPI+0x0C);
    }

    spin_unlock_irqrestore(&Dev->PCI.Interrupt_SpinLock,Flags);

    for(i=0;i<10;i++)
    {
        up(&Dev->PCI.Interrupt_Semaphore);
    }

    return 0;
}



#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,37)
long WRG3_Ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
#else
int WRG3_Ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
#endif
{
 PWRG3 Dev;
 int Error;
 __u8 Value8;

    Dev=(PWRG3)filp->private_data;

    if(Dev==NULL)
    {
        return -EINVAL;
    }


    if(WRG3_LOCK_INTERRUPTIBLE(Dev->Ioctl_Lock))
    {
        return -ERESTARTSYS;
    }

    if(cmd==WRG3_IOCTL_IS_CONNECTED)
    {
        if(Dev->Connected)
        {
            Value8=1;
        }
        else
        {
            Value8=0;
        }

        put_user(Value8,(__u8 __user*)arg);

        Error=0;


        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return Error;
    }
    else
    {
        if(!Dev->Connected)
        {
            WRG3_UNLOCK(Dev->Ioctl_Lock);
            return -EIO;
        }
    }


    switch(cmd)
    {
        case WRG3_IOCTL_PCI_READ_PIC_BYTE:
            Error=WRG3PCI_ReadPICByte(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_WRITE_PIC_BYTE:
            Error=WRG3PCI_WritePICByte(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_READ_PIC_BYTES:
            Error=WRG3PCI_ReadPICBytes(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_WRITE_PIC_BYTES:
            Error=WRG3PCI_WritePICBytes(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_WAIT_FOR_INTERRUPT:
            return WRG3PCI_WaitForInterrupt(Dev,arg);

        case WRG3_IOCTL_USB_GET_COMPLETE_ASYNC_MESSAGE:
            return WRG3USB_GetCompleteAsyncUrb(Dev,arg);

        case WRG3_IOCTL_USB_ISOC_MESSAGE_RECEIVE_ASYNC:
            Error=WRG3USB_IsocMessage_Receive_Async(Dev,arg);
            break;
        case WRG3_IOCTL_USB_BULK_MESSAGE_SEND_ASYNC:
            Error=WRG3USB_BulkMessage_Send_Async(Dev,arg);
            break;
        case WRG3_IOCTL_USB_BULK_MESSAGE_RECEIVE_ASYNC:
            Error=WRG3USB_BulkMessage_Receive_Async(Dev,arg);
            break;
        case WRG3_IOCTL_USB_BULK_MESSAGE_SEND:
            Error=WRG3USB_BulkMessage_Send(Dev,arg);
            break;
        case WRG3_IOCTL_USB_BULK_MESSAGE_RECEIVE:
            Error=WRG3USB_BulkMessage_Receive(Dev,arg);
            break;

        case WRG3_IOCTL_PCI_WRITE:
            Error=WRG3PCI_Write(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_READ:
            Error=WRG3PCI_Read(Dev,arg);
            break;

        case WRG3_IOCTL_USB_CONTROL_MESSAGE:
            Error=WRG3USB_ControlMessage(Dev,arg);
            break;
        case WRG3_IOCTL_USB_CONTROL_MESSAGE_SEND:
            Error=WRG3USB_ControlMessage_Send(Dev,arg);
            break;
        case WRG3_IOCTL_USB_CONTROL_MESSAGE_RECEIVE:
            Error=WRG3USB_ControlMessage_Receive(Dev,arg);
            break;
        case WRG3_IOCTL_USB_CONFIGURE:
            if(!get_user(Value8,(__u8 __user*)arg))
            {
                Error=WRG3USB_Configure(Dev,Value8,1);
            }
            else
            {
                Error=-EFAULT;
            }
            break;
        case WRG3_IOCTL_USB_GET_ENDPOINTINFO:
            Error=WRG3USB_GetEndPointInfo(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_START_INTERRUPT:
            Error=WRG3PCI_StartInterrupt(Dev);
            break;
        case WRG3_IOCTL_PCI_STOP_INTERRUPT:
            Error=WRG3PCI_StopInterrupt(Dev);
            break;
        case WRG3_IOCTL_GET_INTERFACE:
            Error=put_user(Dev->InterfaceType,(__u8 __user*)arg);
            break;
        case WRG3_IOCTL_PCI_RESET:
            Error=WRG3PCI_Reset(Dev);
            break;

        case WRG3_IOCTL_USB_START_ASYNC_MESSAGES:
            Error=WRG3USB_StartAsyncMessages(Dev);
            break;
        case WRG3_IOCTL_USB_STOP_ASYNC_MESSAGES:
            Error=WRG3USB_StopAsyncMessages(Dev);
            break;

        default:
            Error=-EINVAL;
            break;
    }

    WRG3_UNLOCK(Dev->Ioctl_Lock);
 
    return Error;
}

#ifdef CONFIG_COMPAT

#pragma pack(push,1)

typedef struct
{
    __u16           ByteCount;
    compat_uptr_t   Addresses;
    compat_uptr_t   Values;
} COMPAT_PIC_REG;

typedef struct
{
    __u8                    Request;
    __u16                   Value;
    __u16                   Index;
    compat_uptr_t           Buffer;
    __u16                   Size;
    int                     Timeout;
} COMPAT_USB_CONTROL_MSG;

typedef struct
{
    __u8                    EndPoint;
    compat_uptr_t           Buffer;
    __u32                   BufferSize;
    __u32                   Transferred;
    __u32                   Timeout;
} COMPAT_USB_BULK_MESSAGE;

typedef struct
{
    __u8                    EndPoint;
    compat_uptr_t           Buffer;
    __u32                   BufferSize;
    compat_uptr_t           CompletionRoutine;
    compat_uptr_t           Data;
} COMPAT_USB_ASYNC_MESSAGE;

typedef struct
{
    compat_uptr_t           CompletionRoutine;
    compat_uptr_t           Data;
    compat_uptr_t           Buffer;
    __u32                   BufferSize;
    __u32                   Transferred;
    __s32                   Status;
} COMPAT_USB_COMPLETE_ASYNC_MESSAGE;

typedef struct
{
    __u32         Address;
    __u32         Count;
    compat_uptr_t Buffer;
} COMPAT_READ_WRITE_INFO;

#pragma pack(pop)

int WRG3PCI_Compat_ReadPICBytes(PWRG3 Dev,unsigned long arg)
{
 int Error;
 COMPAT_PIC_REG Reg;
 __u16 i;
 __u8 *Addresses;
 __u8 *Values;
 __u8 AddressesBuff[256];
 __u8 ValuesBuff[256];
 __u8 *Address;
 __u8 *Value;
 
    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    } 
 
    Error=0;


    if(copy_from_user(&Reg,(void __user*)arg,sizeof(Reg)))
    {
        return -EFAULT;
    }

    if(Reg.ByteCount==0)
    {
        return 0;
    }


    if(Reg.ByteCount>sizeof(AddressesBuff))
    {
        Addresses=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Addresses==NULL)
        {
            return -ENOMEM;
        }

        Values=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Values==NULL)
        {
            kfree(Addresses);
            return -ENOMEM;
        }
    }
    else
    {
        Addresses=AddressesBuff;
        Values=ValuesBuff;
    }

    if(copy_from_user(Addresses,compat_ptr(Reg.Addresses),Reg.ByteCount)==0)
    {
        Address=Addresses;
        Value=Values;

        for(i=0;i<Reg.ByteCount;i++,Address++,Value++)
        {
            Error=WRG3PCI_ReadPIC(Dev,*Address,Value);

            if(Error!=0)
            {
                break;
            }
        }

        if(Error==0)
        {
            if(copy_to_user(compat_ptr(Reg.Values),Values,Reg.ByteCount))
            {
                Error=-EFAULT;
            }
        }
    }
    else
    {
        Error=-EFAULT;
    }

    if(Addresses!=AddressesBuff)
    {
        kfree(Addresses);
        kfree(Values);
    }

    return Error;
}

int WRG3PCI_Compat_WritePICBytes(PWRG3 Dev,unsigned long arg)
{
 int Error;
 COMPAT_PIC_REG Reg;
 __u16 i;
 __u8 *Addresses;
 __u8 *Values;
 __u8 AddressesBuff[256];
 __u8 ValuesBuff[256];
 __u8 *Address;
 __u8 *Value;

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    } 

    Error=0;

    if(copy_from_user(&Reg,(void __user*)arg,sizeof(Reg)))
    {
        return -EFAULT;
    }

    if(Reg.ByteCount==0)
    {
        return 0;
    }


    if(Reg.ByteCount>sizeof(AddressesBuff))
    {
        Addresses=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Addresses==NULL)
        {
            return -ENOMEM;
        }

        Values=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Values==NULL)
        {
            kfree(Addresses);
            return -ENOMEM;
        }
    }
    else
    {
        Addresses=AddressesBuff;
        Values=ValuesBuff;
    }

    if(copy_from_user(Addresses,compat_ptr(Reg.Addresses),Reg.ByteCount)==0 &&
       copy_from_user(Values,compat_ptr(Reg.Values),Reg.ByteCount)==0)
    {
        Address=Addresses;
        Value=Values;

        for(i=0;i<Reg.ByteCount;i++,Address++,Value++)
        {
            Error=WRG3PCI_WritePIC(Dev,*Address,*Value);

            if(Error!=0)
            {
                break;
            }
        }
    }
    else
    {
        Error=-EFAULT;
    }

    if(Addresses!=AddressesBuff)
    {
        kfree(Addresses);
        kfree(Values);
    }

    return Error;
}


int WRG3USB_Compat_GetCompleteAsyncUrb(PWRG3 Dev,unsigned long arg)
{
 COMPAT_USB_COMPLETE_ASYNC_MESSAGE Result;
 struct list_head *ListEntry;
 PASYNC_URB AsyncUrb;
 unsigned long Flags;
 __u8 *Buffer;
 compat_uptr_t UserBuffer;
 int i;
 struct urb *Urb;

    if(!Dev->Connected || !Dev->USB.AsyncUrbs_Active)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EINVAL;
    }

    spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);

    while(list_empty(&Dev->USB.AsyncUrbs_Complete))
    {
        spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);

        WRG3_UNLOCK(Dev->Ioctl_Lock);

        if(down_interruptible(&Dev->USB.AsyncUrbs_Semaphore))
        {
            return -ERESTARTSYS;
        }

        if(!Dev->Connected)
        {
            return -EINVAL;
        }

        if(!Dev->USB.AsyncUrbs_Active)
        {
            return -EINVAL;
        }

        if(WRG3_LOCK_INTERRUPTIBLE(Dev->Ioctl_Lock))
        {
            return -ERESTARTSYS;
        }


        spin_lock_irqsave(&Dev->USB.AsyncUrbs_SpinLock,Flags);
    }

    ListEntry=Dev->USB.AsyncUrbs_Complete.next;
    AsyncUrb=list_entry(ListEntry,ASYNC_URB,List);
    list_del(ListEntry);

    spin_unlock_irqrestore(&Dev->USB.AsyncUrbs_SpinLock,Flags);


    Urb=AsyncUrb->Urb;

    Result.CompletionRoutine=ptr_to_compat(AsyncUrb->UserCompletionRoutine);
    Result.Buffer=ptr_to_compat(AsyncUrb->UserBuffer);
    Result.BufferSize=AsyncUrb->UserBufferSize;
    Result.Transferred=Urb->actual_length;
    Result.Status=Urb->status;
    Result.Data=ptr_to_compat(AsyncUrb->UserData);

    if(Urb->transfer_buffer!=NULL)
    {
        if(Result.Buffer!=0 && Result.Status==0 && usb_pipein(Urb->pipe))
        {
            if(usb_pipeisoc(Urb->pipe))
            {
                Result.Transferred=0;
                UserBuffer=Result.Buffer;
                Buffer=(__u8*)Urb->transfer_buffer;

                for(i=0;i<Urb->number_of_packets;i++)
                {
                    if(Urb->iso_frame_desc[i].status==0 && Urb->iso_frame_desc[i].actual_length>0)
                    {
                        if(copy_to_user(compat_ptr(UserBuffer),Buffer+Urb->iso_frame_desc[i].offset,Urb->iso_frame_desc[i].actual_length))
                        {
                            Result.Status=-EFAULT;
                            Result.Transferred=0;
                            break;
                        }
                        UserBuffer+=Urb->iso_frame_desc[i].actual_length;
                        Result.Transferred+=Urb->iso_frame_desc[i].actual_length;
                    }
                }
            }
            else
            {
                if(copy_to_user(compat_ptr(Result.Buffer),Urb->transfer_buffer,Urb->actual_length))
                {
                    Result.Status=-EFAULT;
                    Result.Transferred=0;
                }
            }
        }

        kfree(Urb->transfer_buffer);
    }

    usb_free_urb(Urb);

    kfree(AsyncUrb);

    if(copy_to_user((void __user*)arg,&Result,sizeof(Result)))
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EFAULT;
    }
    else
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);

        return 0;
    }
}

int WRG3USB_Compat_BulkMessage_Send_Async(PWRG3 Dev,unsigned long arg)
{
 COMPAT_USB_ASYNC_MESSAGE CompatMsg;
 USB_ASYNC_MESSAGE Msg;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(copy_from_user(&CompatMsg,(void __user*)arg,sizeof(CompatMsg)))
    {
        return -EFAULT;
    }

    Msg.EndPoint=CompatMsg.EndPoint;
    Msg.Buffer=compat_ptr(CompatMsg.Buffer);
    Msg.BufferSize=CompatMsg.BufferSize;
    Msg.CompletionRoutine=compat_ptr(CompatMsg.CompletionRoutine);
    Msg.Data=compat_ptr(CompatMsg.Data);

    return WRG3USB_BulkMessage_Send_Async2(Dev,&Msg);
}

int WRG3USB_Compat_IsocMessage_Receive_Async(PWRG3 Dev,unsigned long arg)
{
 COMPAT_USB_ASYNC_MESSAGE CompatMsg;
 USB_ASYNC_MESSAGE Msg;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(copy_from_user(&CompatMsg,(void __user*)arg,sizeof(CompatMsg)))
    {
        return -EFAULT;
    }


    Msg.EndPoint=CompatMsg.EndPoint;
    Msg.Buffer=compat_ptr(CompatMsg.Buffer);
    Msg.BufferSize=CompatMsg.BufferSize;
    Msg.CompletionRoutine=compat_ptr(CompatMsg.CompletionRoutine);
    Msg.Data=compat_ptr(CompatMsg.Data);

    return WRG3USB_IsocMessage_Receive_Async2(Dev,&Msg);
}

int WRG3USB_Compat_BulkMessage_Receive_Async(PWRG3 Dev,unsigned long arg)
{
 COMPAT_USB_ASYNC_MESSAGE CompatMsg;
 USB_ASYNC_MESSAGE Msg;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(copy_from_user(&CompatMsg,(void __user*)arg,sizeof(CompatMsg)))
    {
        return -EFAULT;
    }

    Msg.EndPoint=CompatMsg.EndPoint;
    Msg.Buffer=compat_ptr(CompatMsg.Buffer);
    Msg.BufferSize=CompatMsg.BufferSize;
    Msg.CompletionRoutine=compat_ptr(CompatMsg.CompletionRoutine);
    Msg.Data=compat_ptr(CompatMsg.Data);

    return WRG3USB_BulkMessage_Receive_Async2(Dev,&Msg);
}

int WRG3USB_Compat_BulkMessage_Send(PWRG3 Dev,unsigned long arg)
{
 COMPAT_USB_BULK_MESSAGE Msg;
 __u8 *Buff;
 int Error;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(copy_from_user(&Msg,(void __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    if(Msg.EndPoint>15 || Dev->USB.USBDevice->ep_out[Msg.EndPoint]==NULL || Msg.Buffer==0 || Msg.BufferSize==0)
    {
        return -EINVAL;
    }

    Buff=(__u8*)kmalloc(Msg.BufferSize,GFP_KERNEL);

    if(Buff==NULL)
    {
        return -ENOMEM;
    }

    if(copy_from_user(Buff,compat_ptr(Msg.Buffer),Msg.BufferSize))
    {
        Error=-EFAULT;
    }
    else
    {
        Error=usb_bulk_msg(Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg.EndPoint),
                           Buff,Msg.BufferSize,&Msg.Transferred,TO_JIFFIES(Msg.Timeout));

        if(Error==-EPIPE)
        {
            usb_clear_halt(Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg.EndPoint));

            Error=usb_bulk_msg(Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg.EndPoint),
                               Buff,Msg.BufferSize,&Msg.Transferred,TO_JIFFIES(Msg.Timeout));
        }

        if(Error==0)
        {
            if(copy_to_user((void __user*)arg,&Msg,sizeof(Msg)))
            {
                Error=-EFAULT;
            }
        }
        else
        {
            if(Error==-EPIPE)
            {
                usb_clear_halt(Dev->USB.USBDevice,usb_sndbulkpipe(Dev->USB.USBDevice,Msg.EndPoint));
            }
        }
    }

    kfree(Buff);

    return Error;
}

int WRG3USB_Compat_BulkMessage_Receive(PWRG3 Dev,unsigned long arg)
{
 COMPAT_USB_BULK_MESSAGE Msg;
 __u8 *Buff;
 int Error;
 
    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }
 

    if(copy_from_user(&Msg,(void __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    if(Msg.EndPoint>15 || Dev->USB.USBDevice->ep_in[Msg.EndPoint]==NULL || Msg.Buffer==0 || Msg.BufferSize==0)
    {
        return -EINVAL;
    }

    Buff=(__u8*)kmalloc(Msg.BufferSize,GFP_KERNEL);

    if(Buff==NULL)
    {
        return -ENOMEM;
    }


    Error=usb_bulk_msg(Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg.EndPoint),
                       Buff,Msg.BufferSize,&Msg.Transferred,TO_JIFFIES(Msg.Timeout));

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg.EndPoint));

        Error=usb_bulk_msg(Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg.EndPoint),
                           Buff,Msg.BufferSize,&Msg.Transferred,TO_JIFFIES(Msg.Timeout));
    }

    if(Error==0)
    {
        if(copy_to_user((void __user*)arg,&Msg,sizeof(Msg)))
        {
            Error=-EFAULT;
        }

        if(copy_to_user(compat_ptr(Msg.Buffer),Buff,Msg.BufferSize))
        {
            Error=-EFAULT;
        }
    }
    else
    {
        if(Error==-EPIPE)
        {
            usb_clear_halt(Dev->USB.USBDevice,usb_rcvbulkpipe(Dev->USB.USBDevice,Msg.EndPoint));
        }
    }

    kfree(Buff);

    return Error;
}

int WRG3PCI_Compat_Write(PWRG3 Dev,unsigned long arg)
{
 COMPAT_READ_WRITE_INFO CompatInfo;
 READ_WRITE_INFO Info;


    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    if(copy_from_user(&CompatInfo,(void __user*)arg,sizeof(CompatInfo)))
    {
        return -EFAULT;
    }

    Info.Address=CompatInfo.Address;
    Info.Count=CompatInfo.Count;
    Info.Buffer=compat_ptr(CompatInfo.Buffer);

    return WRG3PCI_Write2(Dev,&Info);
}

int WRG3PCI_Compat_Read(PWRG3 Dev,unsigned long arg)
{
 COMPAT_READ_WRITE_INFO CompatInfo;
 READ_WRITE_INFO Info;

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    if(copy_from_user(&CompatInfo,(void __user*)arg,sizeof(CompatInfo)))
    {
        return -EFAULT;
    }

    Info.Address=CompatInfo.Address;
    Info.Count=CompatInfo.Count;
    Info.Buffer=compat_ptr(CompatInfo.Buffer);

    return WRG3PCI_Read2(Dev,&Info);
}

int WRG3USB_Compat_ControlMessage_Send(PWRG3 Dev,unsigned long arg)
{
 int Error;
 void *Buff;
 COMPAT_USB_CONTROL_MSG UsbMsg;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -ENOTTY;
    }

    if(copy_from_user(&UsbMsg,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }


    if(UsbMsg.Buffer!=0 && UsbMsg.Size!=0)
    {
        Buff=kmalloc(UsbMsg.Size,GFP_KERNEL);

        if(Buff==NULL)
        {
            return -ENOMEM;
        }

        if(copy_from_user(Buff,compat_ptr(UsbMsg.Buffer),UsbMsg.Size))
        {
            kfree(Buff);
            return -EFAULT;
        }
    }
    else
    {
        Buff=NULL;
        UsbMsg.Size=0;
    }

    Error=usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),
                          UsbMsg.Request,USB_TYPE_VENDOR,
                          UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0));
        usb_clear_halt(Dev->USB.USBDevice,usb_rcvctrlpipe(Dev->USB.USBDevice,0));

        Error=usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),
                          UsbMsg.Request,USB_TYPE_VENDOR,
                          UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));
    }

    if(Error>=0)
    {
        if(Error==UsbMsg.Size)
        {
            Error=0;
        }
        else
        {
            Error=-EIO;
        }
    }

    if(Buff!=NULL)
    {
        kfree(Buff);
    }

    return Error;
}


int WRG3USB_Compat_ControlMessage(PWRG3 Dev,unsigned long arg)
{
 COMPAT_USB_CONTROL_MSG UsbMsg;
 int Error;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -ENOTTY;
    }

    if(copy_from_user(&UsbMsg,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    Error=usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),
                UsbMsg.Request,USB_TYPE_VENDOR,
                UsbMsg.Value,UsbMsg.Index,NULL,0,TO_JIFFIES(UsbMsg.Timeout));

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0));
        usb_clear_halt(Dev->USB.USBDevice,usb_rcvctrlpipe(Dev->USB.USBDevice,0));

        Error=usb_control_msg(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0),
                              UsbMsg.Request,USB_TYPE_VENDOR,
                              UsbMsg.Value,UsbMsg.Index,NULL,0,TO_JIFFIES(UsbMsg.Timeout));
    }

    if(Error>=0)
    {
        Error=0;
    }

 
    return Error;
}

int WRG3USB_Compat_ControlMessage_Receive(PWRG3 Dev,unsigned long arg)
{
 COMPAT_USB_CONTROL_MSG UsbMsg;
 void *Buff;
 int Error;

 
    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -ENOTTY;
    }

    if(copy_from_user(&UsbMsg,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    if(UsbMsg.Buffer!=0 && UsbMsg.Size!=0)
    {
        Buff=kmalloc(UsbMsg.Size,GFP_KERNEL);

        if(Buff==NULL)
        {
            return -ENOMEM;
        }
    }
    else
    {
        Buff=NULL;
        UsbMsg.Size=0;
    }

    Error=usb_control_msg(Dev->USB.USBDevice,usb_rcvctrlpipe(Dev->USB.USBDevice,0),
                          UsbMsg.Request,USB_DIR_IN | USB_TYPE_VENDOR,
                          UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USB.USBDevice,usb_sndctrlpipe(Dev->USB.USBDevice,0));
        usb_clear_halt(Dev->USB.USBDevice,usb_rcvctrlpipe(Dev->USB.USBDevice,0));

        Error=usb_control_msg(Dev->USB.USBDevice,usb_rcvctrlpipe(Dev->USB.USBDevice,0),
                          UsbMsg.Request,USB_DIR_IN | USB_TYPE_VENDOR,
                          UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));
    }

    if(Buff!=NULL)
    {
        if(Error>0)
        {
            if(copy_to_user(compat_ptr(UsbMsg.Buffer),Buff,Error))//UsbMsg.Size))
            {
                Error=-EFAULT;
            }
        }

        kfree(Buff);
    }

    return Error;
}

#define WRG3_COMPAT_IOCTL_PCI_READ_PIC_BYTES                    _IOWR(WRG3_MAGIC,0x06,COMPAT_PIC_REG)
#define WRG3_COMPAT_IOCTL_PCI_WRITE_PIC_BYTES                   _IOW(WRG3_MAGIC,0x07,COMPAT_PIC_REG)
#define WRG3_COMPAT_IOCTL_PCI_WRITE                             _IOW(WRG3_MAGIC,0x0A,COMPAT_READ_WRITE_INFO)
#define WRG3_COMPAT_IOCTL_PCI_READ                              _IOWR(WRG3_MAGIC,0x0B,COMPAT_READ_WRITE_INFO)

#define WRG3_COMPAT_IOCTL_USB_CONTROL_MESSAGE_SEND              _IOW(WRG3_MAGIC,0x20,COMPAT_USB_CONTROL_MSG)
#define WRG3_COMPAT_IOCTL_USB_CONTROL_MESSAGE_RECEIVE           _IOWR(WRG3_MAGIC,0x21,COMPAT_USB_CONTROL_MSG)
#define WRG3_COMPAT_IOCTL_USB_CONTROL_MESSAGE                   _IOW(WRG3_MAGIC,0x22,COMPAT_USB_CONTROL_MSG)
#define WRG3_COMPAT_IOCTL_USB_BULK_MESSAGE_SEND                 _IOWR(WRG3_MAGIC,0x23,COMPAT_USB_BULK_MESSAGE)
#define WRG3_COMPAT_IOCTL_USB_BULK_MESSAGE_SEND_ASYNC           _IOW(WRG3_MAGIC,0x24,COMPAT_USB_ASYNC_MESSAGE)
#define WRG3_COMPAT_IOCTL_USB_BULK_MESSAGE_RECEIVE              _IOWR(WRG3_MAGIC,0x25,COMPAT_USB_BULK_MESSAGE)
#define WRG3_COMPAT_IOCTL_USB_BULK_MESSAGE_RECEIVE_ASYNC        _IOW(WRG3_MAGIC,0x26,COMPAT_USB_ASYNC_MESSAGE)
#define WRG3_COMPAT_IOCTL_USB_GET_COMPLETE_ASYNC_MESSAGE        _IOR(WRG3_MAGIC,0x27,COMPAT_USB_COMPLETE_ASYNC_MESSAGE)
#define WRG3_COMPAT_IOCTL_USB_ISOC_MESSAGE_RECEIVE_ASYNC        _IOW(WRG3_MAGIC,0x2C,COMPAT_USB_ASYNC_MESSAGE)

long WRG3_Compat_Ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
 PWRG3 Dev;
 int Error;
 u_int8_t Value8;

    Dev=(PWRG3)filp->private_data;

    if(Dev==NULL)
    {
        return -EINVAL;
    }


    if(WRG3_LOCK_INTERRUPTIBLE(Dev->Ioctl_Lock))
    {
        return -ERESTARTSYS;
    }

    if(cmd==WRG3_IOCTL_IS_CONNECTED)
    {
        if(Dev->Connected)
        {
            Value8=1;
        }
        else
        {
            Value8=0;
        }

        Error=put_user(Value8,(__u8 __user*)arg);

        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return Error;
    }
    else
    {
        if(!Dev->Connected)
        {
            WRG3_UNLOCK(Dev->Ioctl_Lock);
            return -EIO;
        }
    }


    switch(cmd)
    {
        case WRG3_IOCTL_PCI_READ_PIC_BYTE:
            Error=WRG3PCI_ReadPICByte(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_WRITE_PIC_BYTE:
            Error=WRG3PCI_WritePICByte(Dev,arg);
            break;
        case WRG3_COMPAT_IOCTL_PCI_READ_PIC_BYTES:
            Error=WRG3PCI_Compat_ReadPICBytes(Dev,arg);
            break;
        case WRG3_COMPAT_IOCTL_PCI_WRITE_PIC_BYTES:
            Error=WRG3PCI_Compat_WritePICBytes(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_WAIT_FOR_INTERRUPT:
            return WRG3PCI_WaitForInterrupt(Dev,arg);

        case WRG3_COMPAT_IOCTL_USB_GET_COMPLETE_ASYNC_MESSAGE:
            return WRG3USB_Compat_GetCompleteAsyncUrb(Dev,arg);

        case WRG3_COMPAT_IOCTL_USB_ISOC_MESSAGE_RECEIVE_ASYNC:
            Error=WRG3USB_Compat_IsocMessage_Receive_Async(Dev,arg);
            break;
        case WRG3_COMPAT_IOCTL_USB_BULK_MESSAGE_SEND_ASYNC:
            Error=WRG3USB_Compat_BulkMessage_Send_Async(Dev,arg);
            break;
        case WRG3_COMPAT_IOCTL_USB_BULK_MESSAGE_RECEIVE_ASYNC:
            Error=WRG3USB_Compat_BulkMessage_Receive_Async(Dev,arg);
            break;
        case WRG3_COMPAT_IOCTL_USB_BULK_MESSAGE_SEND:
            Error=WRG3USB_Compat_BulkMessage_Send(Dev,arg);
            break;

        case WRG3_COMPAT_IOCTL_USB_BULK_MESSAGE_RECEIVE:
            Error=WRG3USB_Compat_BulkMessage_Receive(Dev,arg);
            break;

        case WRG3_COMPAT_IOCTL_PCI_WRITE:
            Error=WRG3PCI_Compat_Write(Dev,arg);
            break;

        case WRG3_COMPAT_IOCTL_PCI_READ:
            Error=WRG3PCI_Compat_Read(Dev,arg);
            break;

        case WRG3_COMPAT_IOCTL_USB_CONTROL_MESSAGE:
            Error=WRG3USB_Compat_ControlMessage(Dev,arg);
            break;
        case WRG3_COMPAT_IOCTL_USB_CONTROL_MESSAGE_SEND:
            Error=WRG3USB_Compat_ControlMessage_Send(Dev,arg);
            break;
        case WRG3_COMPAT_IOCTL_USB_CONTROL_MESSAGE_RECEIVE:
            Error=WRG3USB_Compat_ControlMessage_Receive(Dev,arg);
            break;
        case WRG3_IOCTL_USB_CONFIGURE:
            if(!get_user(Value8,(__u8 __user*)arg))
            {
                Error=WRG3USB_Configure(Dev,Value8,1);
            }
            else
            {
                Error=-EFAULT;
            }
            break;

        case WRG3_IOCTL_USB_GET_ENDPOINTINFO:
            Error=WRG3USB_GetEndPointInfo(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_START_INTERRUPT:
            Error=WRG3PCI_StartInterrupt(Dev);
            break;
        case WRG3_IOCTL_PCI_STOP_INTERRUPT:
            Error=WRG3PCI_StopInterrupt(Dev);
            break;
        case WRG3_IOCTL_GET_INTERFACE:
            Error=put_user(Dev->InterfaceType,(__u8 __user*)arg);
            break;
        case WRG3_IOCTL_PCI_RESET:
            Error=WRG3PCI_Reset(Dev);
            break;

        case WRG3_IOCTL_USB_START_ASYNC_MESSAGES:
            Error=WRG3USB_StartAsyncMessages(Dev);
            break;
        case WRG3_IOCTL_USB_STOP_ASYNC_MESSAGES:
            Error=WRG3USB_StopAsyncMessages(Dev);
            break;

        default:
            Error=-EINVAL;
            break;
    }

    WRG3_UNLOCK(Dev->Ioctl_Lock);
 
    return Error;
}

#endif 

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,26)
static void WRG3ReleaseCD(struct device *cd)
#else
static void WRG3ReleaseCD(struct class_device *cd)
#endif
{
}


int RegisterWRG3Device(PWRG3 Dev)
{
 int i;
 int DeviceIndex;
 int Minor;
 dev_t DevNo;
 int Error;

    DeviceIndex=-1;


    if(WRG3_LOCK_INTERRUPTIBLE(WRG3_Device_Lock))
    {
        return -ERESTARTSYS;
    }

    if(Dev->DeviceIndex>=0)
    {
        //device is already registered

        WRG3_UNLOCK(WRG3_Device_Lock);
        return 0;
    }

    for(i=0;i<MAX_WRG3_DEVICES;i++)
    {
        if(WRG3_Devices[i]==NULL)
        {
            DeviceIndex=i;
            break;
        }
    }

    if(DeviceIndex<0)
    {
        WRG3_UNLOCK(WRG3_Device_Lock);
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Maximum number of devices [%d] exceeded.\n",MAX_WRG3_DEVICES);
        return -ENFILE;
    }

    Minor=WRG3_MINOR_BASE+DeviceIndex;

    DevNo=MKDEV(WiNRADiO_MAJOR,Minor);

    Dev->Minor=Minor;
    Dev->DeviceIndex=DeviceIndex;

    memset(&Dev->ClassDevice,0,sizeof(Dev->ClassDevice));


    Dev->ClassDevice.class=&WRG3_Class;
    Dev->ClassDevice.devt=MKDEV(WiNRADiO_MAJOR,Minor);

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,26)
    dev_set_name(&Dev->ClassDevice,WRG3_NAME"%d",DeviceIndex);
#else
    snprintf(Dev->ClassDevice.class_id,BUS_ID_SIZE,WRG3_NAME"%d",DeviceIndex);
#endif

    Error=device_register(&Dev->ClassDevice);

    if(Error<0)
    {
        WRG3_UNLOCK(WRG3_Device_Lock);
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to register class (%d).\n",Error);
        return Error;
    }

    cdev_init(&Dev->cdev,&WRG3_File_Operations);
    Dev->cdev.owner=THIS_MODULE;
    Dev->cdev.ops=&WRG3_File_Operations;
    Error=cdev_add(&Dev->cdev,DevNo,1);

    if(Error<0)
    {
        device_unregister(&Dev->ClassDevice);

        WRG3_UNLOCK(WRG3_Device_Lock);
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to add character device (%d).\n",Error);
        return Error;
    }

    WRG3_Devices[DeviceIndex]=Dev;

    WRG3_UNLOCK(WRG3_Device_Lock);

    return 0;
}

int UnregisterWRG3Device(PWRG3 Dev)
{
    if(WRG3_LOCK_INTERRUPTIBLE(WRG3_Device_Lock))
    {
        return -ERESTARTSYS;
    }

    if(Dev->DeviceIndex>=0 && Dev->DeviceIndex<MAX_WRG3_DEVICES)
    {
        if(WRG3_Devices[Dev->DeviceIndex]==Dev)
        {
            WRG3_Devices[Dev->DeviceIndex]=NULL;
            cdev_del(&Dev->cdev);

            device_unregister(&Dev->ClassDevice);
        }
    }

    WRG3_UNLOCK(WRG3_Device_Lock);

    return 0;
}

module_init(WRG3_Init);
module_exit(WRG3_Exit);

MODULE_LICENSE("GPL");

//EXPORT_SYMBOL_GPL();
